@@ -1,13 +1,43 @@
-#!/usr/bin/perl
-#
-# viperdb.pl - Filesystem Integrity Monitor
-#
-# Copyright (C) 1998-2001 J-Dog <J-Dog@Resentment.org> & Peter Surda <shurdeek@panorama.sth.ac.at>
-#
-# This program is free software; you can redistribute it and/or
-# modify it under the terms of the GNU General Public License
-# as published by the Free Software Foundation; either version 2
-# of the License, or (at your option) any later version.
+#!/usr/bin/perl -w
+
+# ViperDB v0.9.9
+
+# (c) j-dog@resentment.org, shurdeek@panorama.sth.ac.at and many others 
+# 1998-2001
+# licensed under GNU GPL v2
+
+# ViperDB was created as a smaller & faster option to Tripwire.  Tripwire while
+# being a great product leaves something to be desired in the speed department
+# and also, by default tripwire generates a report everytime it runs and
+# directs that report to an email address. This hinders most people from
+# running Tripwire every few minutes to do a system check. ViperDB however is
+# the answer to this problem. ViperDB does not use a fancy all-in-one database
+# to keep records instead, I opted to keep it fast and hence decided to go with
+# a plaintext db which is stored in each "watched" directory.
+
+# UPADATE shurdeek 27.2.2001 - Now it uses GDBM, but is still in each watched
+# directory
+
+# By using this there is no real one attack point for a attacker to focus his
+# attention on. This coupled with the running of ViperDB every 5 minutes (via
+# cron root job) decreases that likelyhood that an attacker will be able to
+# modify your "watched" filesystem while ViperDB is monitoring your system
+
+# PLANNED UPGRADES: (by the original authors I'm unable to contact)
+# - Adding of a more complex "system status" function which when
+#   a change is detected, would grab info that might be helpful
+#   in determining what caused the change (ie. processes running,
+#   users logged in, last few lines from logfiles, etc)
+
+# THANKS TO:
+# whitetrash, punkis, & rooster, and those of you who
+# have sent in your suggestions.
+# UPDATE by shurdeek: thanx to j-dog :-)
+
+# Special thanks to Anthony D. Urso <anthonyu@killa.net> for
+# finding the nasty race conditions which prompted me to recode
+# all file access properly instead of being lame and using system()
+# for every file access
 
 use MD5;
 use GDBM_File;
@@ -21,7 +51,6 @@ my $configfile = '/etc/viperdb.conf';
 my $dbfname = '.ViperDB';
 my $lock = "/var/lock/subsys/viperdb.lock";
 my $md5lock = "/var/lock/subsys/viperdb-md5.lock";
-my $tielimit = 5;
 my $strict = 0;
 my $init = 0;
 my $doattr = 0;
@@ -43,7 +72,7 @@ chomp($os_type=lc(qx[uname]));
 
 # logfile is only used if defined in config file.
 if (defined $logfile) {
-	open LOG, ">>$logfile";
+	open LOG, ">>$logfile" or do { warn "$!: $logfile"; undef $logfile; };
 }
 
 if ($ARGV[1] eq '-md5') {
@@ -61,8 +90,9 @@ if ($ARGV[0] eq '-init'){
 	print "Check Detected: Now Checking File Sanity...\n";
 	&Go;
 } else {
-	print "\n\nViperDB v0.9.8\n";
+	print "\n\nViperDB v0.9.5\n";
 	print "ERROR: Unrecognized option or none given.\n";
+	print "Homepage http://panorama.sth.ac.at/viperdb/\n";
 	print "usage: ViperDB -init|-check|-checkstrict [-md5]\n";
 	print "     -init	  Initializes the ViperDB Databases\n";
 	print "     -check	Runs a system file sanity check\n";
@@ -81,7 +111,8 @@ sub Lock
 	if (-e $lock) {
 		open LOCK, $lock;
 		$tmp = <LOCK>;
-		$pid = chomp $tmp;
+		chomp $tmp;
+		$pid = $tmp;
 		close LOCK;
 		@tmp = `ps $pid|grep $pid`;
 		if (scalar @tmp == 1 || time () - $mtime < $timelock) {
@@ -118,7 +149,7 @@ sub ParseConfigs
 	open (CONFIG, "< $configfile");
 	while (<CONFIG>) {
 		chomp;
-		if (/\[(.+)\]/) {
+		if (/^\[(.+)\]/) {
 			$sub = $1;
 		# empty, comment or whitespace-only
 		} elsif (! defined $_ or /(^\s*(#|:|;))|(^\s*$)/ ) {
@@ -159,8 +190,6 @@ sub ParseConfigs
 					$md5lock = $1;
 				} elsif (/^timelock\s*=\s*(.+)/) {
 					$timelock = $1;
-				} elsif (/^tielimit\s*=\s*(.+)/) {
-					$tielimit = $1;
 				} elsif (/^dbfname\s*=\s*(.+)/) {
 					$dbfname = $1;
 				} elsif (/^logfile\s*=\s*(.+)/) {
@@ -171,7 +200,7 @@ sub ParseConfigs
 	}
 
 	$found = 1;
-	%ndirs = {};
+	%ndirs = ();
 	my $file;
 	while ($found) {
 		$found = 0;
@@ -206,7 +235,7 @@ sub ParseConfigs
 				}
 			closedir(DIRLIST);
 			%rdirs = %ndirs; # copy list of new directories
-			%ndirs = {}; # reset
+			%ndirs = (); # reset
 		}
 	}
 
@@ -233,7 +262,7 @@ sub Err
 	}
 
 	my ($tmp) = sprintf "%-15s: %s\n", uc $class, $err;
-	printf LOG $tmp;
+	printf LOG $tmp if (defined $logfile);
 	$errorsummary .= $tmp;
 }
 
@@ -271,52 +300,33 @@ sub OpenDB
 		}
 	}
 	# create or open.
-	$dbopened = 1;
 	if ($strict) {
 		$numtries = 0;
-		while (++$numtries <= $tielimit) {
-			untie %db;
-			tie (%db, 'GDBM_File', $fname, &GDBM_READER, 0400) && last;
-		}
-		if ($numtries != 1) {
-			if ($numtries >= $tielimit) {
-				$dbopened = 0;
-				&Err ("critical", 
-					"In $wd, can't tie $dbfname after $numtries tries.\nDatabase probably corrupt, suggesting deleting.");
-			} else {
-				&Err ("suspicious", 
-					"In $wd, tie to $dbfname needed $numtries tries.");
-			}
+		untie %db;
+		if (! tie (%db, 'GDBM_File', $fname, &GDBM_READER, 0400)) {
+			$dbopened = 0;
+			&Err ("critical", 
+				"In $wd, can't tie $dbfname.\nDatabase probably corrupt, suggesting deleting.");
+		} else {
+			$dbopened = 1;
 		}
 	} else {
 		# if exists, chmod so we can write
 		( -f $fname ) && chmod (0600, $fname);
 		$numtries = 0;
-		while (++$numtries <= $tielimit) {
 			untie %db;
-			tie (%db, 'GDBM_File', $fname, &GDBM_WRCREAT, 0600) && last;
-		}
-		if ($numtries != 1) {
-			if ($numtries >= $tielimit) {
+		if (! tie (%db, 'GDBM_File', $fname, &GDBM_WRCREAT, 0600)) {
+			$dbopened = 0;
+			&Err ("critical", 
+				"In $wd, can't tie $dbfname.\nDeleting and retrying...");
+			unlink $fname;
+			untie %db;
+			if (! tie (%db, 'GDBM_File', $fname, &GDBM_WRCREAT, 0600)) {
 				$dbopened = 0;
-				&Err ("critical", 
-					"In $wd, can't tie $dbfname after $numtries tries.\nDeleting and retrying...");
-				$numtries = 0;
-				unlink $fname;
-				while (++$numtries <= $tielimit) {
-					untie %db;
-					tie (%db, 'GDBM_File', $fname, &GDBM_WRCREAT, 0600) && last;
-				}
-				if ($numtries >= $tielimit) {
-					$dbopened = 0;
-					&Err ("critical", "Failed. Unable to monitor $wd.");
-				} else {
-					&Err ("note", "Success after $numtries tries.");
-				}
-			} else {
-				&Err ("suspicious", 
-					"In $wd, tie to $dbfname needed $numtries tries.");
+				&Err ("critical", "Failed. Unable to monitor $wd.");
 			}
+		} else {
+			$dbopened = 1;
 		}
 	}
 }
@@ -561,7 +571,7 @@ sub Go
 		$str = &strftime ("%c", localtime);
 		&Err ("info", "$str\nEND RUN - $trouble changed files detected.");
 	}
-	close (LOG);
+	close (LOG) if (defined $logfile);
 	if (defined $email && $trouble && ! $init) {
 		open (MAIL, "|mail -s '$subject' $email");
 			print MAIL "$errorsummary";
